/****************************************************************************

    DRC: Digital Room Correction
    Copyright (C) 2002 Denis Sbragion

		This file is part of DRC, Digital Room Correction

		DRC  is  distributed  with  NO WARRANTY OF ANY KIND. No author or
		distributor accepts any responsibility for  the  consequences  of
		using  it,  or  for  whether  it serves any particular purpose or
		works at all, unless he or she says so in writing.  Refer to  the
		Aladdin Free Public License (the "License") for full details.

		Every copy of DRC must include a copy of the License, normally in
		a plain ASCII text file named PUBLIC. The License grants you  the
		right  to  copy, modify and redistribute DRC, but only under cer
		tain conditions described in the License.   Among  other  things,
		the License requires that the copyright notice and this notice be
		preserved on all copies.

		You can contact the author on Internet at the following address:

				d.sbragion@infotecna.it

		This program uses the parsecfg library from Yuuki  NINOMIYA.  De
		tails  on  this  library  can be found in the parsecfg.c and par
		secfg.h files.  Many thanks to Yuuki NINOMIYA for this useful li
		brary.

****************************************************************************/

/* Funzioni di libreria base */

/* Inclusioni */
#include "baselib.h"
#include "fft.h"
#include <stdio.h>
#include <math.h>

/* Output stringhe con sync output */
int sputs(const char * s)
	{
		int Res;
		Res = puts(s);
		fflush(stdout);
		return(Res);
	}

/* Determina la lunghezza di un file */
size_t FSize(FILE * F)
	{
		long CPos;
		long FS;

		CPos = ftell(F);
		fseek(F,0,SEEK_END);
		FS = ftell(F);
		fseek(F,CPos,SEEK_SET);
		return (size_t) FS;
	}

/* Legge parte di un file di ingresso e lo pone nell'array indicato 
	di dimensione InitWindow. 
	Ritorna true se l'operazione ha successo.
*/
Boolean ReadSignal(const char * FName,DRCFloat * Dst,const int InitWindow,
	const int ImpulseCenter,const IFileType FType)
	{
		int IBS;
		int I;
		int IStart;
		int IEnd;
		DRCFileFloat RWF;
		DRCFileInt RWI;
		int RWDim;

		/* File gestione IO */
		FILE * IOF;
		long FS;

		/* Azzera l'array destinazione */
		for (I = 0;I < InitWindow;I++)
			Dst[I] = 0;

		/* Salva la dimensione del blocco lettura */
		switch (FType)
			{
				case PcmFloat32Bit:
					RWDim = sizeof(DRCFileFloat);
				break;

				case PcmInt16Bit:
					RWDim = sizeof(DRCFileInt);
				break;
			}

		/* Apre il file di input */
		if ((IOF = fopen(FName,"rb")) == NULL)
			{
				perror("\nUnable to open input file.");
				return False;
			}

		/* Recupera la dimensione del file */
		FS = FSize(IOF) / RWDim;
		
		if (FS == 0)
			{
				puts("\nInput file is zero length.");
				return False;
			}

		if (ImpulseCenter > FS)
			{
				puts("\nImpulseCenter is out of input file range.");
				return False;
			}

		/* Calcola la dimensione del blocco in ingresso */		
		if (InitWindow == 0)
			if (ImpulseCenter >= FS / 2)
				IBS = 1 + ImpulseCenter * 2;
			else
				IBS = ((FS - ImpulseCenter) * 2) - 1;
		else
			IBS = InitWindow;

		/* Salta il segmento iniziale */
		if (ImpulseCenter > InitWindow / 2)
			{
				if (fseek(IOF,(ImpulseCenter - InitWindow / 2) * RWDim,SEEK_SET) != 0)
					{
						perror("\nError reading input file.");
						return False;
					}
				IStart = 0;
				IEnd = FS - (ImpulseCenter - InitWindow / 2);
			}
		else
			{
				IStart = InitWindow / 2 - ImpulseCenter;
				IEnd = FS;
			}

		/* Legge la risposta all'impulso */
		switch (FType)
			{
				case PcmFloat32Bit:
					for (I = IStart; I < InitWindow && I < IEnd; I++)
						{	
							if (fread((void *) &RWF,sizeof(DRCFileFloat),1,IOF) != 1)
								{
									perror("\nError reading input file.");
									return False;
								}
							Dst[I] = (DLReal) RWF;
						}
				break;

				case PcmInt16Bit:
					for (I = IStart; I < InitWindow && I < IEnd; I++)
						{	
							if (fread((void *) &RWI,sizeof(DRCFileInt),1,IOF) != 1)
								{
									perror("\nError reading input file.");
									return False;
								}
							Dst[I] = (DLReal) RWI;
						}
				break;
			}

		/* Chiude il file */
		fclose(IOF);

		/* Operazione completata */
		return True;
	}

/* Scrive il segnale indicato su disco */
Boolean WriteSignal(const char * FName,const DRCFloat * Src,const int SSize,
	const IFileType FType)
	{
		/* File gestione IO */
		FILE * IOF;
		int I;
		DRCFileFloat RWF;
		DRCFileInt RWI;

		/* Apre il file di output */
		IOF = fopen(FName,"wb");

		/* Salva la risposta risultante */
		switch (FType)
			{
				case PcmFloat32Bit:
					for (I = 0; I < SSize; I++)
						{
							RWF = (DRCFileFloat) Src[I];
							if (fwrite((void *) &RWF,sizeof(DRCFileFloat),1,IOF) != 1)
								{
									perror("\nError writing ouput file.");
									return False;
								}
						}
				break;

				case PcmInt16Bit:
					for (I = 0; I < SSize; I++)
						{
							RWI = (DRCFileInt) floor(0.5 + Src[I]);
							if (fwrite((void *) &RWI,sizeof(DRCFileInt),1,IOF) != 1)
								{
									perror("\nError writing ouput file.");
									return False;
								}
						}
				break;
			}

		/* Chiude il file */
		fclose(IOF);

		/* Operazione completata */
		return True;
	}

/* Calcola l'autocorrelazione del sgnale S */
Boolean AutoCorrelation(DLReal * S, int N)
	{
		DLComplex * C;
		int I;

		if ((C = new(DLComplex[N])) == NULL)
			return False;

		for (I = 0; I < N; I++)
			C[I] = S[I];

		Fft(C,N);
		
		for (I = 0; I < N; I++)
			C[I] = C[I] * std::conj(C[I]);
			
		IFft(C,N);
		
		for (I = 0; I < N; I++)
			S[I] = std::real(C[I]) / N;

		delete C;

		return True;
	}

/* Legge i punti di generazione filtro FIR dal file indicato */
Boolean ReadPoints(char * CorrFile,DLReal * FilterFreqs,
	DLComplex * FilterPoints,int NPoints,int SampleRate)
	{
		/* File gestione IO */
		FILE * IOF;

		/* Variabili ausiliarie */
		int I;
		float Freq;
		float Mag;

		/* Apre il file di correzione */
		if ((IOF = fopen(CorrFile,"rt")) == NULL)
			{
				perror("\nUnable to open correction file.");
				return False;
			}
	
		/* Legge i valori */
		for (I = 0; I < NPoints; I++)
			{
				if (fscanf(IOF,"%f %f",&Freq,&Mag) != 2)
					{
						perror("\nError reading correction file.");
						return False;
					}
				FilterFreqs[I] = (DLReal) (2 * Freq) / SampleRate;
				FilterPoints[I] = (DLReal) Mag;
			}
		FilterFreqs[NPoints - 1] = 1;

		/* Operazione completata */
		return True;

	}

/* Trova il valore massimo all'interno di un file. Versione float 32 bit. */
static int FindMaxPcmFloat32Bit(const char * FName)
	{
		DRCFileFloat RWF;
		int MaxPos;
		int Pos;
		DRCFileFloat PcmMax;

		/* File gestione IO */
		FILE * IOF;

		/* Apre il file di input */
		if ((IOF = fopen(FName,"rb")) == NULL)
			{
				perror("\nUnable to open input file.");
				return -1;
			}

		/* Ciclo ricerca massimo */
		Pos = 0;
		MaxPos = 0;
		PcmMax = 0;
		while(feof(IOF) == 0)
			{
				if (fread((void *) &RWF,sizeof(DRCFileFloat),1,IOF) != 1)
					{
						if (feof(IOF) == 0)
							{
								perror("\nError reading input file.");
								return -1;
							}
					}

				if (fabs(RWF) > PcmMax)
					{
						PcmMax = RWF;
						MaxPos = Pos;
					}

				Pos++;
			}

		/* Operazione completata */
		return MaxPos;
	}

/* Trova il valore massimo all'interno di un file. Versione int 16 Bit. */
static int FindMaxPcmInt16Bit(const char * FName)
	{
		DRCFileInt RWI;
		int MaxPos;
		int Pos;
		DRCFileInt PcmMax;

		/* File gestione IO */
		FILE * IOF;

		/* Apre il file di input */
		if ((IOF = fopen(FName,"rb")) == NULL)
			{
				perror("\nUnable to open input file.");
				return -1;
			}

		/* Ciclo ricerca massimo */
		Pos = 0;
		MaxPos = 0;
		PcmMax = 0;
		while(feof(IOF) == 0)
			{
				if (fread((void *) &RWI,sizeof(DRCFileInt),1,IOF) != 1)
					{
						if (feof(IOF) == 0)
							{
								perror("\nError reading input file.");
								return -1;
							}
					}

				if (abs(RWI) > PcmMax)
					{
						PcmMax = RWI;
						MaxPos = Pos;
					}

				Pos++;
			}

		/* Operazione completata */
		return MaxPos;
	}

/* Trova il valore massimo all'interno di un file. */
int FindMaxPcm(const char * FName,const IFileType FType)
	{
		/* Controlla il tipo file */
		switch (FType)
			{
				case PcmFloat32Bit:
					return FindMaxPcmFloat32Bit(FName);

				case PcmInt16Bit:
					return FindMaxPcmInt16Bit(FName);
			}

		/* Tipo file errato */
		return -1;
	}
