/****************************************************************************

    DRC: Digital Room Correction
    Copyright (C) 2002 Denis Sbragion

		This file is part of DRC, Digital Room Correction

		DRC  is  distributed  with  NO WARRANTY OF ANY KIND. No author or
		distributor accepts any responsibility for  the  consequences  of
		using  it,  or  for  whether  it serves any particular purpose or
		works at all, unless he or she says so in writing.  Refer to  the
		Aladdin Free Public License (the "License") for full details.

		Every copy of DRC must include a copy of the License, normally in
		a plain ASCII text file named PUBLIC. The License grants you  the
		right  to  copy, modify and redistribute DRC, but only under cer
		tain conditions described in the License.   Among  other  things,
		the License requires that the copyright notice and this notice be
		preserved on all copies.

		You can contact the author on Internet at the following address:

				d.sbragion@infotecna.it

		This program uses the parsecfg library from Yuuki  NINOMIYA.  De
		tails  on  this  library  can be found in the parsecfg.c and par
		secfg.h files.  Many thanks to Yuuki NINOMIYA for this useful li
		brary.

****************************************************************************/

/****************************************************************************
  Progetto    : DSP Library.
  File        : Fft.cpp
  Autore      : Sbragion Denis
  Descrizione : Funzioni per fft, ifft etc.
  Revisioni   :
  16/10/93    : Prima stesura.
****************************************************************************/

#include "fft.h"

DLComplex Power(DLComplex X,unsigned int N)
  {
    unsigned int I;
    DLComplex Pow(1);

    for(I = 1;I <= N;I <<= 1)
      {
        if ((I & N) != 0)
          Pow *= X;
        X *= X;
      }

    return(Pow);
  }

DLComplex UnitRoot(unsigned int I,unsigned int N)
  {
    DLReal Arg = (DLReal) (((2*M_PI)*(I%N))/N);
    return(DLComplex((DLReal) cos(Arg),(DLReal) sin(Arg)));
  }

unsigned int FirstFactor(unsigned int N)
  {
    unsigned int I;

    if (N < 2)
      return(1);
    else
      if (N%2 == 0)
        return(2);
      else
        {
          for(I = 3;I <= N/I && N%I != 0; I += 2);

          if (I > N/I)
            I = N;
        }

    return(I);
  }

unsigned int Radix2Reverse(unsigned int I,unsigned int N)
  {
    unsigned int IR,J,Bit;

    IR = 0;

    for(J = 1,Bit = N >> 1;J < N;J <<= 1,Bit >>= 1)
      if((I & J) != 0)
        IR |= Bit;

    return(IR);
  }

void Radix2Shuffle(DLComplex P[],unsigned int N)
  {
    unsigned int I,IR;
    DLComplex Temp;

    for(I = 0;I < N;I++)
      {
        IR = Radix2Reverse(I,N);

        if(I < IR)
          {
            Temp = P[I];
            P[I] = P[IR];
            P[IR] = Temp;
          }
      }
  }

void Radix2Fft(DLComplex P[],unsigned int N)
  {
    unsigned int I,J,K,Span,Start,End;
    DLComplex P0,Root;

    for(K = 0,Span = N >> 1;K < Span;K++)
      {
        P0 = P[Span+K];
        P[Span+K] = P[K]-P0;
        P[K] += P0;
      }

    for(I = 2,Span >>= 1;I < N;I <<= 1,Span >>= 1)
      {
        for(J = 0;J < I;J++)
          {
						Root = std::conj(UnitRoot(J,I << 1));

            Start = (Radix2Reverse(J,I) << 1)*Span;
            End = Start+Span;

            for(K = Start;K < End;K++)
              {
                P0 = Root*P[Span+K];
                P[Span+K] = P[K]-P0;
                P[K] += P0;
              }
          }
      }

    Radix2Shuffle(P,N);
  }

void Radix2IFft(DLComplex P[],unsigned int N)
  {
    unsigned int I,J,K,Span,Start,End;
    DLComplex P0,Root;

    for(K = 0,Span = N >> 1;K < Span;K++)
      {
        P0 = P[Span+K];
        P[Span+K] = P[K]-P0;
        P[K] += P0;
      }

    for(I = 2,Span >>= 1;I < N;I <<= 1,Span >>= 1)
      {
        for(J = 0;J < I;J++)
          {
            Root = UnitRoot(J,I << 1);

            Start = (Radix2Reverse(J,I) << 1)*Span;
            End = Start+Span;

            for(K = Start;K < End;K++)
              {
                P0 = Root*P[Span+K];
                P[Span+K] = P[K]-P0;
                P[K] += P0;
              }
          }
      }

    Radix2Shuffle(P,N);

    for(I = 0;I < N;I++)
      P[I] /= (DLReal) N;
  }

unsigned int MixedReverse(unsigned int I,unsigned int N)
  {
    unsigned int IR,J,Digit,Radix;

    if (N < 2)
      return(I);
    else
      {
        IR = 0;

        Radix = FirstFactor(N);

        for(J = 1,Digit = N/Radix;J < N;Digit /= Radix)
          {
            IR += ((I/J)%Radix)*Digit;

            J *= Radix;
            Radix = FirstFactor(Digit);
          }
      }

    return(IR);
  }

void MixedShuffle(DLComplex P[],unsigned int N)
  {
    unsigned int I,IR,J;
    DLComplex Temp;

    for(I = 1;I < N-1;I++)
      {
        IR = MixedReverse(I,N);
        while(I > IR)
          IR = MixedReverse(IR,N);

        if (I == IR)
          {
            J = I;
            IR = MixedReverse(I,N);
            Temp = P[J];
            while(I != IR)
              {
                P[J] = P[IR];
                J = IR;
                IR = MixedReverse(IR,N);
              }
            P[J] = Temp;
          }
      }
  }

Boolean Fft(DLComplex P[],unsigned int N)
  {
    unsigned int I,J,K,L,Span,Radix,Start,End,RP;
    DLComplex * PT;
    DLComplex Root,P0;

    Radix = FirstFactor(N);

    for(I = 1,Span = N/Radix;I < N;Span /= Radix)
      {
        RP = I*Radix;

        if (Radix == 2)
          if (I == 1)
            for(K = 0;K < Span;K++)
              {
                P0 = P[Span+K];
                P[Span+K] = P[K]-P0;
                P[K] += P0;
              }
          else
            for(J = 0;J < I;J++)
              {
                Start = MixedReverse(J,I)*Radix*Span;
                End = Start+Span;
                Root = std::conj(UnitRoot(J,RP));

                for(K = Start;K < End;K++)
                  {
                    P0 = Root*P[Span+K];
                    P[Span+K] = P[K]-P0;
                    P[K] += P0;
                  }
              }
        else
          if ((PT = new(DLComplex[Radix])) != NULL)
            {
              for(J = 0;J < I;J++)
                {
                  Start = MixedReverse(J,I)*Radix*Span;
                  End = Start+Span;

                  for(K = Start;K < End;K++)
                    {
                      for (L = 0;L < Radix;L++)
                        PT[L] = P[Span*L+K];

                      for (L = 0;L < Radix;L++)
                        P[Span*L+K] = PolyEval(PT,Radix,std::conj(UnitRoot(L*I+J,RP)));
                    }
                }

              delete(PT);
            }
          else
            return(False);

        I *= Radix;
        Radix = FirstFactor(Span);
      }

    MixedShuffle(P,N);

    return(True);
  }

Boolean IFft(DLComplex P[],unsigned int N)
  {
    unsigned int I,J,K,L,Span,Radix,Start,End,RP;
    DLComplex * PT;
    DLComplex Root,P0;

    Radix = FirstFactor(N);

    for(I = 1,Span = N/Radix;I < N;Span /= Radix)
      {
        RP = I*Radix;

        if (Radix == 2)
          if (I == 1)
            for(K = 0;K < Span;K++)
              {
                P0 = P[Span+K];
                P[Span+K] = P[K]-P0;
                P[K] += P0;
              }
          else
            for(J = 0;J < I;J++)
              {
                Start = MixedReverse(J,I)*Radix*Span;
                End = Start+Span;
                Root = UnitRoot(J,RP);

                for(K = Start;K < End;K++)
                  {
                    P0 = Root*P[Span+K];
                    P[Span+K] = P[K]-P0;
                    P[K] += P0;
                  }
              }
        else
          if ((PT = new(DLComplex[Radix])) != NULL)
            {
              for(J = 0;J < I;J++)
                {
                  Start = MixedReverse(J,I)*Radix*Span;
                  End = Start+Span;

                  for(K = Start;K < End;K++)
                    {
                      for (L = 0;L < Radix;L++)
                        PT[L] = P[Span*L+K];

                      for (L = 0;L < Radix;L++)
                        P[Span*L+K] = PolyEval(PT,Radix,
                          UnitRoot(L*I+J,RP));
                    }
                }

              delete(PT);
            }
          else
            return(False);

        I *= Radix;
        Radix = FirstFactor(Span);
      }

    MixedShuffle(P,N);

    for(I = 0;I < N;I++)
      P[I] /= (DLReal) N;

    return(True);
  }

DLComplex PolyEval(DLComplex P[],unsigned int N,DLComplex X)
  {
    DLComplex Y(0);
    unsigned int I;

    if (N > 0)
      {
        Y = P[N-1];
        for(I = N-1;I > 0;I--)
          Y = X*Y+P[I-1];
      }

    return(Y);
  }

Boolean Dft(DLComplex P[],unsigned int N)
  {
    unsigned int I;
    DLComplex * R;

    if ((R = new(DLComplex[N])) != NULL)
      {
        for(I = 0;I < N;I++)
          R[I] = PolyEval(P,N,std::conj(UnitRoot(I,N)));

        for(I = 0;I < N;I++)
          P[I] = R[I];

        delete(R);

        return(True);
      }
    else
      return(False);
  }

Boolean IDft(DLComplex P[],unsigned int N)
  {
    unsigned int I;
    DLComplex * R;

    if ((R = new(DLComplex[N])) != NULL)
      {
        for(I = 0;I < N;I++)
          R[I] = PolyEval(P,N,UnitRoot(I,N));

        for(I = 0;I < N;I++)
          P[I] = R[I]/((DLReal) N);

        delete(R);

        return(True);
      }
    else
      return(False);
  }

/***************************************************************************/
