/****************************************************************************

    DRC: Digital Room Correction
    Copyright (C) 2002 Denis Sbragion

		This file is part of DRC, Digital Room Correction

		DRC  is  distributed  with  NO WARRANTY OF ANY KIND. No author or
		distributor accepts any responsibility for  the  consequences  of
		using  it,  or  for  whether  it serves any particular purpose or
		works at all, unless he or she says so in writing.  Refer to  the
		Aladdin Free Public License (the "License") for full details.

		Every copy of DRC must include a copy of the License, normally in
		a plain ASCII text file named PUBLIC. The License grants you  the
		right  to  copy, modify and redistribute DRC, but only under cer
		tain conditions described in the License.   Among  other  things,
		the License requires that the copyright notice and this notice be
		preserved on all copies.

		You can contact the author on Internet at the following address:

				d.sbragion@infotecna.it

		This program uses the parsecfg library from Yuuki  NINOMIYA.  De
		tails  on  this  library  can be found in the parsecfg.c and par
		secfg.h files.  Many thanks to Yuuki NINOMIYA for this useful li
		brary.

****************************************************************************/

/****************************************************************************
  Progetto    : DSP Library.
  File        : Fir.cpp
  Autore      : Sbragion Denis
  Descrizione : Funzioni per il calcolo di filtri fir.
  Revisioni   :
  16/10/93    : Prima stesura.
****************************************************************************/

#include "fir.h"
#include "fft.h"
#include <stddef.h>

void LowPassFir(DLReal * Filter,unsigned int Order,DLReal Freq)
  {
    unsigned int I,HalfOrder = Order/2;
    DLReal C = (DLReal) (M_PI*Freq);

    if (Order > 1)
      if (Order%2 == 0)
        for(I = 1; I <= HalfOrder;I++)
          {
            Filter[HalfOrder-I] = (DLReal) (sin((I-0.5)*C)/((I-0.5)*M_PI));
            Filter[HalfOrder+I-1] = Filter[HalfOrder-I];
          }
      else
        {
          for(I = 1; I <= HalfOrder;I++)
            {
              Filter[HalfOrder-I] = (DLReal) (sin(I*C)/(I*M_PI));
              Filter[HalfOrder+I] = Filter[HalfOrder-I];
            }
          Filter[HalfOrder] = Freq;
        }
  }

void HighPassFir(DLReal * Filter,unsigned int Order,DLReal Freq)
  {
    unsigned int I,HalfOrder;
    DLReal C = (DLReal) (M_PI*Freq);

    if (Order > 2)
      {
        if (Order%2 == 0)
          {
            Order--;
            Filter[Order] = 0;
          }

        HalfOrder = Order/2;

        for(I = 1; I <= HalfOrder;I++)
          {
            Filter[HalfOrder-I] = (DLReal) (-sin(I*C)/(I*M_PI));
            Filter[HalfOrder+I] = Filter[HalfOrder-I];
          }
        Filter[HalfOrder] = 1-Freq;
      }
  }

void BandPassFir(DLReal * Filter,unsigned int Order,DLReal Low,DLReal High)
  {
    unsigned int I,HalfOrder = Order/2;
    DLReal D = (DLReal) (M_PI*(High-Low)/2), S = (DLReal) (M_PI*(High+Low)/2);

    if (Order > 1)
      if (Order%2 == 0)
        for(I = 1; I <= HalfOrder;I++)
          {
            Filter[HalfOrder-I] = (DLReal) (M_2_PI*sin((I-0.5)*D)*
              cos((I-0.5)*S)/(I-0.5));
            Filter[HalfOrder+I-1] = Filter[HalfOrder-I];
          }
      else
        {
          for(I = 1; I <= HalfOrder;I++)
            {
              Filter[HalfOrder-I] = (DLReal) (M_2_PI*sin(I*D)*
                cos(I*S)/I);
              Filter[HalfOrder+I] = Filter[HalfOrder-I];
            }
          Filter[HalfOrder] = M_PI*M_PI*(High-Low);
        }
  }

void BandStopFir(DLReal * Filter,unsigned int Order,DLReal Low,DLReal High)
  {
    unsigned int I,HalfOrder;
    DLReal D = (DLReal) (M_PI*(High-Low)/2),S = (DLReal) (M_PI*(High+Low)/2);

    if (Order > 2)
      {
        if (Order%2 == 0)
          {
            Order--;
            Filter[Order] = 0;
          }

        HalfOrder = Order/2;

        for(I = 1; I <= HalfOrder;I++)
          {
            Filter[HalfOrder-I] = (DLReal) (-M_2_PI*sin(I*D)*
              cos(I*S)/I);
            Filter[HalfOrder+I] = Filter[HalfOrder-I];
          }
        Filter[HalfOrder] = 1-(High-Low);
      }
  }

void HilbertFir(DLReal * Filter,unsigned int Order)
  {
    unsigned int I,HalfOrder;

    if (Order > 2)
      {
        if (Order%2 == 0)
          {
            Order--;
            Filter[Order] = 0;
          }

        HalfOrder = Order/2;

        for(I = 1;I <= HalfOrder;I++)
          {
            Filter[HalfOrder+I] = (DLReal) (M_2_PI*(I%2)/I);
            Filter[HalfOrder-I] = -Filter[HalfOrder+I];
          }

        Filter[HalfOrder] = 0;
      }
  }

void Interpolate(DLComplex * A,unsigned int Size,DLReal * X,
  DLComplex * Y)
  {
    unsigned int I,P;
    DLReal Mag,DeltaMag,Arg,DeltaArg;
    DLReal StopI,Last,Pc;

    P = 0;
    StopI = 0;

    for(I = 0;I < Size;)
      {
        if (I >= StopI)
          {
						Mag = std::abs(Y[P]);						
						DeltaMag = std::abs(Y[P+1])-Mag;
            Arg = std::arg(Y[P]);
            DeltaArg = std::arg(Y[P+1])-Arg;
            Last = StopI;
            StopI = X[P+1]*Size;
            P++;
          }
        else
          {
            Pc = (I-Last)/(StopI-Last);
            A[I] = std::polar(Mag+DeltaMag*Pc,Arg+DeltaArg*Pc);
            I++;
          }
      }
  }

Boolean GenericFir(DLReal * Filter,unsigned int Order,DLReal * F,
  DLComplex * G,unsigned int Np,unsigned int Is)
  {
    DLComplex * Ht;
    unsigned int I,HalfFilter,HtSize,HtSize2;

    if ((Order < 2) || (Np < 2))
      return(False);

    if ((F[0] != 0) || (F[Np-1] != 1))
      return(False);

    for(I = 0;I < Np-1;I++)
      if (F[I] > F[I+1])
        return(False);

    if (Is == 0)
      for (HtSize = 2;HtSize < Order;HtSize <<= 1);
    else
      HtSize = Is;

    if (HtSize < Order)
      return(False);

    HtSize2 = 2*HtSize;

    if ((Ht = new(DLComplex[HtSize2])) == NULL)
      return(False);

    Interpolate(Ht,HtSize,F,G);

    if (Order%2 == 0)
      {
        for(I = 1;I < HtSize;I++)
          Ht[HtSize2-I] = -std::conj(Ht[I]);
        Ht[HtSize] = G[Np-1];
        for (I = 0;I < HtSize2;I++)
          Ht[I] *= UnitRoot(I,2*HtSize2);
      }
    else
      {
        for(I = 1;I < HtSize;I++)
          Ht[2*HtSize-I] = std::conj(Ht[I]);
        Ht[HtSize] = G[Np-1];
      }

    if (IFft(Ht,HtSize2) == False)
      {
        delete(Ht);
        return(False);
      }

    HalfFilter = Order/2;

    for(I = 0;I < HalfFilter;I++)
      Filter[I] = std::real(Ht[HtSize2-HalfFilter+I]);
    for(I = HalfFilter;I < Order;I++)
      Filter[I] = std::real(Ht[I-HalfFilter]);

    delete(Ht);
    return(True);
  }

/***************************************************************************/
