/****************************************************************************

    DRC: Digital Room Correction
    Copyright (C) 2002 Denis Sbragion

		This file is part of DRC, Digital Room Correction

		DRC  is  distributed  with  NO WARRANTY OF ANY KIND. No author or
		distributor accepts any responsibility for  the  consequences  of
		using  it,  or  for  whether  it serves any particular purpose or
		works at all, unless he or she says so in writing.  Refer to  the
		Aladdin Free Public License (the "License") for full details.

		Every copy of DRC must include a copy of the License, normally in
		a plain ASCII text file named PUBLIC. The License grants you  the
		right  to  copy, modify and redistribute DRC, but only under cer
		tain conditions described in the License.   Among  other  things,
		the License requires that the copyright notice and this notice be
		preserved on all copies.

		You can contact the author on Internet at the following address:

				d.sbragion@infotecna.it

		This program uses the parsecfg library from Yuuki  NINOMIYA.  De
		tails  on  this  library  can be found in the parsecfg.c and par
		secfg.h files.  Many thanks to Yuuki NINOMIYA for this useful li
		brary.

****************************************************************************/

/* Funzioni correzione e gestione livello del segnale */

/* Inclusioni */
#include "level.h"
#include "baselib.h"
#include "fft.h"
#include <math.h>

/* Calola il valore RMS del segnale Sig */ 
DLReal GetRMSLevel(const DLReal * Sig,const int SigLen)
	{
		DLReal RMS;
		int I;

		RMS = 0;
		for (I = 0; I < SigLen; I++)	
			RMS += Sig[I] * Sig[I];

		return sqrt(RMS);
	}

/* Effettua la normalizzazione del segnale al valore indicato e
secondo il metodo indicato */
void SigNormalize(DLReal * Sig,const int SigLen,const DLReal NormFactor,
	const NormType TNorm)
	{
		DLReal NormBase;
		int I;

		switch (TNorm)
			{
				case NormEuclidean:
					{
						NormBase = GetRMSLevel(Sig,SigLen);
						if (NormBase > 0)
							for (I = 0; I < SigLen; I++)	
								Sig[I] = (Sig[I] * NormFactor) / NormBase;
					}
				break;

				case NormMax:
					{
						NormBase = 0;
						for (I = 0; I < SigLen; I++)	
							if (((DLReal) std::abs<DLReal>(Sig[I])) > NormBase)
								NormBase = (DLReal) (std::abs<DLReal>(Sig[I]));
				
						for (I = 0; I < SigLen; I++)	
							Sig[I] = (Sig[I] * NormFactor) / NormBase;
					}
				break;
				
				case NormSum:
					{
						NormBase = 0;
						for (I = 0; I < SigLen; I++)	
							NormBase += std::abs<DLReal>(Sig[I]);

						for (I = 0; I < SigLen; I++)	
							Sig[I] = (Sig[I] * NormFactor) / NormBase;
					}
				break;
			}
	}

Boolean DipLimit(DLReal * Sig,const int SigLen,const DLReal MinGain,const int MExp)
	{
		int FFTSize;
		int I;
		DLComplex * FFTArray;
		DLReal RMSLevel;

		/* Determina la prima potenza di 2 >= SigLen */
		for(FFTSize = 1;FFTSize <= SigLen;FFTSize <<= 1);
		FFTSize *= 1 << MExp;

		/* Alloca l'array per l'fft */
		FFTArray = new DLComplex[FFTSize];
		if (FFTArray == NULL)
			return False;

		/* Copia il segnale nell'array */
		for (I = 0;I < SigLen;I++)
			FFTArray[I] = Sig[I];
		for (I = SigLen;I < FFTSize;I++)
			FFTArray[I] = 0;

		/* Determina il livello RMS del segnale */
		RMSLevel = GetRMSLevel(Sig,SigLen);

		/* Effettua l'fft del segnale */
		Radix2Fft(FFTArray,FFTSize);

		/* Scansione per la limitazione guadagno */
		for (I = 0; I < FFTSize; I++)	
			if (std::abs(FFTArray[I]) < RMSLevel * MinGain)
				FFTArray[I] = (RMSLevel * MinGain * FFTArray[I]) / std::abs(FFTArray[I]);

		/* Effettua l'fft inversa del segnale */
		Radix2IFft(FFTArray,FFTSize);

		/* Ritorna il segnale limitato */
		for (I = 0;I < SigLen;I++)
			Sig[I] = std::real(FFTArray[I]);

		/* Dealloca l'array */
		delete FFTArray;

		/* Operazione completata */
		return True;
	}

Boolean PeakLimit(DLReal * Sig,const int SigLen,const DLReal MaxGain,const int MExp)
	{
		int FFTSize;
		int I;
		DLComplex * FFTArray;
		DLReal RMSLevel;

		/* Determina la prima potenza di 2 >= SigLen */
		for(FFTSize = 1;FFTSize <= SigLen;FFTSize <<= 1);
		FFTSize *= 1 << MExp;

		/* Alloca l'array per l'fft */
		FFTArray = new DLComplex[FFTSize];
		if (FFTArray == NULL)
			return False;

		/* Copia il segnale nell'array */
		for (I = 0;I < SigLen;I++)
			FFTArray[I] = Sig[I];
		for (I = SigLen;I < FFTSize;I++)
			FFTArray[I] = 0;

		/* Determina il livello RMS del segnale */
		RMSLevel = GetRMSLevel(Sig,SigLen);

		/* Effettua l'fft del segnale */
		Radix2Fft(FFTArray,FFTSize);

		/* Scansione per la limitazione guadagno */
		for (I = 0; I < FFTSize; I++)	
			if (std::abs(FFTArray[I]) > RMSLevel * MaxGain)
				FFTArray[I] = (RMSLevel * MaxGain * FFTArray[I]) / std::abs(FFTArray[I]);

		/* Effettua l'fft inversa del segnale */
		Radix2IFft(FFTArray,FFTSize);

		/* Ritorna il segnale limitato */
		for (I = 0;I < SigLen;I++)
			Sig[I] = std::real(FFTArray[I]);

		/* Dealloca l'array */
		delete FFTArray;

		/* Operazione completata */
		return True;
	}

/* Livellazione norma */
Boolean NormFlat(DLReal * Sig,const int SigLen,const DLReal Gain,const int MExp)
	{
		int FFTSize;
		int I;
		DLComplex * FFTArray;

		/* Determina la prima potenza di 2 >= SigLen */
		for(FFTSize = 1;FFTSize <= SigLen;FFTSize <<= 1);
		FFTSize *= 1 << MExp;

		/* Alloca l'array per l'fft */
		FFTArray = new DLComplex[FFTSize];
		if (FFTArray == NULL)
			return False;

		/* Copia il segnale nell'array */
		for (I = 0;I < SigLen;I++)
			FFTArray[I] = Sig[I];
		for (I = SigLen;I < FFTSize;I++)
			FFTArray[I] = 0;

		/* Effettua l'fft del segnale */
		Radix2Fft(FFTArray,FFTSize);

		/* Scansione per la limitazione guadagno */
		for (I = 0; I < FFTSize; I++)	
			FFTArray[I] = (Gain * FFTArray[I]) / std::abs(FFTArray[I]);

		/* Effettua l'fft inversa del segnale */
		Radix2IFft(FFTArray,FFTSize);

		/* Ritorna il segnale limitato */
		for (I = 0;I < SigLen;I++)
			Sig[I] = std::real(FFTArray[I]);

		/* Dealloca l'array */
		delete FFTArray;

		/* Operazione completata */
		return True;
	}
